/*
 * Copyright (C) 2003  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 * stats.c -- Handle statistical information, such as load averages.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <assert.h>
#include <stdlib.h>
#include "timewrap.h"
#include <string.h>
#include <stdio.h>
#include <setjmp.h>
#include <float.h>
#include <math.h>

#include "stats.h"
#include "log.h"
#include "lyskomd.h"
#include "kom-types.h"
#include "services.h"
#include "async.h"
#include "com.h"
#include "connections.h"
#include "kom-errno.h"
#include "manipulate.h"
#include "server/smalloc.h"

#define FACT_00 (1)		/*  1 second */
#define FACT_0 (15)		/* 15 seconds */
#define FACT_1 ( 4 * FACT_0)	/*  1 minute */
#define FACT_2 ( 5 * FACT_1)	/*  5 minutes */
#define FACT_3 ( 3 * FACT_2)	/* 15 minutes */
#define N_FACTS 5
#define HISTORY_LENGTH (FACT_3)

/* Define this to debug the average counting code.  This will slow
   down the code a lot, so it is not recommended for production use.  */
#undef DEBUG_STATS

enum value_type { VT_AVG, VT_ASCEND, VT_DESCEND, NUM_VT };

struct avg_status {
    struct timeval when;
    long value;
    double acc;
    float avg_history[HISTORY_LENGTH];
    double avenrun[N_FACTS];
    long updates[N_FACTS];
    float max_seen[N_FACTS];
};

static struct avg_status status[NUM_STAT][NUM_VT];
static struct avg_status copy[NUM_STAT][NUM_VT];

static const int factors[] = {
    FACT_00,
    FACT_0,
    FACT_1,
    FACT_2,
    FACT_3,
};

static int
check_stat(enum stat_type st,
	   const char *when);

static int
ind(unsigned long sec,
    unsigned int offset)
{
    int ret = (sec - offset) % HISTORY_LENGTH;
    assert(ret >= 0);
    assert(ret < HISTORY_LENGTH);
    return ret;
}

    
#ifdef DEBUG_STATS
static void
dump_stats(struct avg_status s[NUM_STAT][NUM_VT])
{
    enum stat_type st;
    enum value_type vt;
    int i;

    fprintf(stderr, "START DUMP\n");
    for (vt = 0; vt < NUM_VT; vt++)
    {
	fprintf(stderr, "value_type %d:\n", (int)vt);
	for (st = 0; st < NUM_STAT; st++)
	{
	    fprintf(stderr, "stat_type %d: when=%ld:%ld value=%ld acc=%f\n",
		    (int)st,
		    (long)s[st][vt].when.tv_sec, (long)s[st][vt].when.tv_usec,
		    s[st][vt].value, s[st][vt].acc);
	    fprintf(stderr, "  indices:");
	    fprintf(stderr, " ():0:%d", ind(s[st][vt].when.tv_sec, 0));
	    for (i = 0; i < N_FACTS; i++)
		fprintf(stderr, " %d:%d:%d", i, factors[i],
			ind(s[st][vt].when.tv_sec, factors[i]));
	    fprintf(stderr, "\n  avenrun:");
	    for (i = 0; i < N_FACTS; i++)
		fprintf(stderr, " %f", s[st][vt].avenrun[i]);
	    fprintf(stderr, "\n  updates:");
	    for (i = 0; i < N_FACTS; i++)
		fprintf(stderr, " %ld", s[st][vt].updates[i]);
	    fprintf(stderr, "\n  max_seen:");
	    for (i = 0; i < N_FACTS; i++)
		fprintf(stderr, " %g", s[st][vt].max_seen[i]);
	    fprintf(stderr, "\n  history:");
	    for (i = 0; i < HISTORY_LENGTH; i++)
		fprintf(stderr, " %f", s[st][vt].avg_history[i]);
	    fprintf(stderr, "\n");
	}
    }
}
#endif

void
init_stats(void)
{
    enum stat_type st;
    enum value_type vt;
    int i;
    struct timeval now;

    assert(sizeof(factors)/sizeof(factors[0]) == N_FACTS);
    assert(factors[N_FACTS-1] == HISTORY_LENGTH);

    gettimeofday(&now, NULL);

    for (st = 0; st < NUM_STAT; st++)
	for (vt = 0; vt < NUM_VT; vt++)
	{
	    status[st][vt].when = now;
	    status[st][vt].value = 0;
	    status[st][vt].acc = 0.0;
	    for (i = 0; i < HISTORY_LENGTH; i++)
		status[st][vt].avg_history[i] = 0.0;
	    for (i = 0; i < N_FACTS; i++)
	    {
		status[st][vt].avenrun[i] = 0.0;
		status[st][vt].updates[i] = 0;
		status[st][vt].max_seen[i] = 0.0;
	    }
	}
}


static void
update_history(struct avg_status *s,
	       time_t end,
	       float value)
{
    int i;

    for (i = 0; i < N_FACTS; i++)
	if (value > s->max_seen[i])
	    s->max_seen[i] = value;

    for (; s->when.tv_sec < end; s->when.tv_sec++)
    {
	for (i = 0; i < N_FACTS; i++)
	{
	    s->avenrun[i] += (value - s->avg_history[ind(s->when.tv_sec,
							 factors[i])])
		/ (double)factors[i];
	    s->updates[i]++;
	}
	
	s->avg_history[ind(s->when.tv_sec, 0)] = value;
    }

}

	       
static void
update_one_stat(enum stat_type st,
		enum value_type vt,
		long delta,
		struct timeval now)
{
    long time_delta;
    int i;
    struct avg_status *s = &status[st][vt];

    if (now.tv_sec == s->when.tv_sec)
    {
	if (vt == VT_AVG)
	{
	    time_delta = now.tv_usec - s->when.tv_usec;
	    s->acc += 1.0e-6 * (double)s->value * time_delta;
	    s->when.tv_usec = now.tv_usec;
	}
    }
    else if (now.tv_sec - s->when.tv_sec <= HISTORY_LENGTH)
    {
	if (vt == VT_AVG)
	{
	    time_delta = 1000000 - s->when.tv_usec;
	    s->acc += 1.0e-6 * (double)s->value * time_delta;
	}
	else
	    s->acc = s->value;

	update_history(s, s->when.tv_sec+1, s->acc);
	if (vt != VT_AVG)
	{
	    s->acc = 0;
	    s->value = 0;
	}
	update_history(s, now.tv_sec, s->value);
	assert(s->when.tv_sec == now.tv_sec);
	if (vt == VT_AVG)
	{
	    s->acc = 1.0e-6 * (double)s->value * now.tv_usec;
	    s->when.tv_usec = now.tv_usec;
	}
    }
    else
    {
	if (vt != VT_AVG)
	{
	    s->acc = 0;
	    s->value = 0;
	}
	for (i = 0; i < N_FACTS; i++)
	{
	    s->avenrun[i] = s->value;
	    s->updates[i] = 0;
	}
	for (i = 0; i < HISTORY_LENGTH; i++)
	    s->avg_history[i] = s->value;
	if (vt == VT_AVG)
	    s->acc = 1.0e-6 * (double)s->value * now.tv_usec;
	s->when = now;
    }

    s->value += delta;
    if (s->value < 0)
    {
	kom_log("update_stat(%d, %ld): current value became negative: %ld\n",
		st, delta, s->value);
	s->value = 0;
    }
#ifdef DEBUG_STATS
    if (check_stat(st, "update_stat") > 0)
    {
	kom_log("This was caused by an update of delta=%ld for type %d\n",
		delta, st);
	kom_log("Dump of before:\n");
	dump_stats(copy);
	kom_log("Dump of after:\n");
	dump_stats(status);
    }
#endif
}

void
update_stat(enum stat_type st,
	    long delta)
{
    struct timeval now;

    memcpy(copy, status, sizeof(status));
    gettimeofday(&now, NULL);

    update_one_stat(st, VT_AVG, delta, now);
    if (delta >= 0)
	update_one_stat(st, VT_ASCEND, delta, now);
    if (delta <= 0)
	update_one_stat(st, VT_DESCEND, -delta, now);
}

static int
is_nonzero(double val)
{
#ifdef FP_ZERO
    return fpclassify(val) != FP_ZERO;
#else
    return val >= DBL_MIN || val <= -DBL_MIN;
#endif
}

static int
check_one_stat(enum stat_type st,
	       enum value_type vt,
	       const char *when)
{
    struct avg_status *s;
    double acc;
    double err;
    float val;
    int f;
    int i;
    int changes = 0;

    /* FIXME: This could be rewritten so that it makes a single pass
       through the avg_history.  There is no real need to scan the
       first parts once for each factor.  */

    s = &status[st][vt];
    for (f = 0; f < N_FACTS; f++)
    {
	acc = 0;
	for (i = 0; i < factors[f]; i++)
	{
	    val = s->avg_history[ind(s->when.tv_sec, i+1)];
	    assert(val >= 0);
	    acc += val;
	}
	acc /= factors[f];
	err = acc - s->avenrun[f];
	if (err < 0)
	    err = -err;
	if ((acc < 1 && err > 1e5 * acc)
	    || (acc >= 1 && err / acc > 1e-5)
	    || s->avenrun[f] < 0)
	{
	    /* Don't bother logging assignments to 0.  */
	    if (is_nonzero(acc) || err > 1e-7 * s->max_seen[f])
	    {
		kom_log("Accumulated rounding errors fixed (%s) after"
			" %ld updates (st=%d, vt=%d, f=%d, factor=%d):"
			" %g => %g\n",
			when,
			s->updates[f], st, vt, f, factors[f],
			s->avenrun[f], acc);
	    }
	    s->avenrun[f] = acc;
	    s->updates[f] = 0;
	    s->max_seen[f] = acc;
	    changes++;
	}
    }
    return changes;
}

static int
check_stat(enum stat_type st,
	   const char *when)
{
    enum value_type vt;
    int changes = 0;

    for (vt = 0; vt < NUM_VT; vt++)
	changes += check_one_stat(st, vt, when);

    return changes;
}


int
check_all_stats(const char *when)
{
    enum stat_type st;
    int changes = 0;

    for (st = 0; st < NUM_STAT; st++)
	changes += check_stat(st, when);

    return changes;
}


static const char *
name(enum stat_type st)
{
    switch (st)
    {
    case STAT_RUN_QUEUE:
	return "run-queue-length";

    case STAT_DNS_QUEUE:
	return "pending-dns";

    case STAT_IDENT_QUEUE:
	return "pending-ident";

    case STAT_CLIENTS:
	return "clients";

    case STAT_REQUESTS:
	return "reqs";

    case STAT_TEXTS:
	return "texts";

    case STAT_CONFS:
	return "confs";

    case STAT_PERSONS:
	return "persons";

    case STAT_SEND_QUEUE:
	return "send-queue-bytes";

    case STAT_RECV_QUEUE:
	return "recv-queue-bytes";

    case NUM_STAT:
	abort();
    }

    abort();
}

long
read_stat_value(enum stat_type st)
{
    return status[st][VT_AVG].value;
}

extern Success
get_stats_description(Stats_description *result)
{
    enum stat_type st;
    int i;

    CHK_CONNECTION(FAILURE);

    result->no_of_stats = NUM_STAT;
    result->stat_names = tmp_alloc(NUM_STAT * sizeof(result->stat_names[0]));

    for (st = 0; st < NUM_STAT; st++)
	result->stat_names[st] = s_fcrea_str(name(st));

    result->intervals.length = N_FACTS + 1;
    result->intervals.data = tmp_alloc(
	(N_FACTS + 1) * sizeof(result->intervals.data[0]));

    result->intervals.data[0] = 0; /* The first momentaneous value. */
    for (i = 0; i < N_FACTS; i++)
	result->intervals.data[i+1] = factors[i];

    return OK;
}

extern Success
get_stats(const String what,
	  Stats_list *result)
{
    enum stat_type st;
    int i;

    CHK_CONNECTION(FAILURE);

    for (st = 0; st < NUM_STAT; st++)
	if (s_streq(what, s_fcrea_str(name(st))))
	    break;

    if (st == NUM_STAT)
    {
	kom_errno = KOM_UNDEFINED_MEASUREMENT;
	err_stat = 0;
	return FAILURE;
    }

    /* Make sure we get the curren statistics, not some old value. */
    update_stat(st, 0);
    check_stat(st, "get_stats");

    result->no_of_stats = N_FACTS + 1;
    result->stats = tmp_alloc((N_FACTS + 1) * sizeof(result->stats[0]));

    result->stats[0].average = status[st][VT_AVG].value;
    result->stats[0].ascent_rate = 0;
    result->stats[0].descent_rate = 0;
    
    for (i = 0; i < N_FACTS; i++)
    {
	result->stats[i+1].average = status[st][VT_AVG].avenrun[i];
	result->stats[i+1].ascent_rate = status[st][VT_ASCEND].avenrun[i];
	result->stats[i+1].descent_rate = status[st][VT_DESCEND].avenrun[i];
    }

    return OK;
}
